# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import argparse
import csv
import logging
import re
import shutil
import warnings
from pathlib import Path
from typing import Sequence
from urllib.parse import urlparse

import click
import fsspec
import numpy as np
import rasterio as rio
import yaml
from fsspec.core import OpenFile
from tqdm.contrib.logging import logging_redirect_tqdm
from tqdm.std import tqdm

from orthority import root_path, utils
from orthority.camera import create_camera, FrameCamera
from orthority.enums import CameraType, Compress, Interp
from orthority.errors import CrsError, CrsMissingError, ParamError
from orthority.factory import Cameras, FrameCameras, RpcCameras
from orthority.ortho import Ortho
from orthority.version import __version__

logger = logging.getLogger(__name__)

# TODO: use universal_pathlib if/when it matures for all filename options instead of OpenFile,
#  and the utils.get_filename and utils.join_ofile work arounds.


class RstCommand(click.Command):
    """click.Command subclass for formatting help with RST markup."""

    def get_help(self, ctx: click.Context):
        """
        Strip some RST markup from the help text for CLI display.

        Doesn't work with grid tables.
        """

        # Note that this can't easily be done in __init__, as each sub-command's __init__ gets
        # called, which ends up re-assigning self.wrap_text to reformat_text
        if not hasattr(self, 'click_wrap_text'):
            self.click_wrap_text = click.formatting.wrap_text

        sub_strings = {
            # convert from RST friendly to click literal (unwrapped) block marker
            '\b\n': '\n\b',
            # strip RST literal (unwrapped) marker in e.g. tables and bullet lists
            r'\| ': '',
            # strip RST ref directive '\n.. _<name>:\n'
            r'\n\.\. _.*:\n': '',
            # convert from RST '::' to ':'
            '::': ':',
            # convert from RST '``literal``' to 'literal'
            '``(.*?)``': r'\g<1>',
            # convert ':option:`--name <group-command --name>`' to '--name'
            r':option:`(.*?)(\s+<.*?>)?`': r'\g<1>',
            # convert ':option:`--name`' to '--name'
            ':option:`(.*?)`': r'\g<1>',
            # convert ':file:`file/na.me`' to "'file/na.me'"
            ':file:`(.*?)`': r"'\g<1>'",
            # strip '----...'
            # '-{4,}': r'',
            # # convert from RST cross-ref '`<name> <link>`__' to 'name'
            # r'`(.*?)(\s+<.*?>)?`_+': r'\g<1>',
            # convert from RST cross-ref '`<name> <link>`__' to 'link'
            r'`(.*?)<(.*?)>`_+': r'\g<2>',
        }

        def reformat_text(text, width, **kwargs):
            for sub_key, sub_value in sub_strings.items():
                text = re.sub(sub_key, sub_value, text, flags=re.DOTALL)
            wr_text = self.click_wrap_text(text, width, **kwargs)
            # change double newline to single newline separated list
            return re.sub('\n\n(\s*?)- ', '\n- ', wr_text, flags=re.DOTALL)

        click.formatting.wrap_text = reformat_text
        return click.Command.get_help(self, ctx)


def _configure_logging(verbosity: int):
    """Configure python logging level."""

    def showwarning(message, category, filename, lineno, file=None, line=None):
        """Redirect orthority warnings to the source module's logger, otherwise show warning as
        usual.
        """
        # adapted from https://discuss.python.org/t/some-easy-and-pythonic-way-to-bind-warnings-to-loggers/14009/2
        package_root = Path(__file__).parents[1]
        file_path = Path(filename)
        try:
            module_path = file_path.relative_to(package_root)
            is_relative_to = True
        except ValueError:
            is_relative_to = False

        if file is not None or not is_relative_to:
            orig_show_warning(message, category, filename, lineno, file, line)
        else:
            module_name = module_path.with_suffix('').as_posix().replace('/', '.')
            logger = logging.getLogger(module_name)
            logger.warning(str(message))

    # redirect orthority warnings to module logger
    orig_show_warning = warnings.showwarning
    warnings.showwarning = showwarning

    # Configure the package logger, leaving logs from dependencies on their defaults (e.g. for
    # rasterio, they stay hidden).
    # Adapted from rasterio: https://github.com/rasterio/rasterio/blob/main/rasterio/rio/main.py.
    log_level = max(10, 20 - 10 * verbosity)
    pkg_logger = logging.getLogger(__package__)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('%(levelname)s:%(name)s: %(message)s'))
    pkg_logger.addHandler(handler)
    pkg_logger.setLevel(log_level)


def _read_crs(crs: str):
    """Read a CRS from a string, text file, or image file."""
    crs_path = Path(crs)
    if crs_path.suffix.lower() in ['.tif', '.tiff']:
        # read CRS from geotiff path / URL
        with utils.suppress_no_georef(), utils.OpenRaster(crs, 'r') as im:
            crs = im.crs
    else:
        if crs_path.exists() or urlparse(crs).scheme in fsspec.available_protocols():
            # read string from text file path / valid URI
            with utils.Open(crs, 'rt') as f:
                crs = f.read()
        # read CRS from string
        crs = rio.CRS.from_string(crs)
    return crs


def _crs_cb(ctx: click.Context, param: click.Parameter, crs: str):
    """Click callback to validate and parse the CRS."""
    if crs is not None:
        try:
            crs = _read_crs(crs)
        except Exception as ex:
            raise click.BadParameter(f'{str(ex)}', param=param)
    return crs


def _src_files_cb(
    ctx: click.Context, param: click.Parameter, src_files: Sequence[str]
) -> Sequence[OpenFile]:
    """Click callback to form a list of source image OpenFile instances. ``src_files`` can be a
    list of file paths / URIs, or a list of path / URI glob patterns.
    """
    if not src_files or len(src_files) == 0:
        return ()
    try:
        src_files = fsspec.open_files(list(src_files), 'rb')
    except Exception as ex:
        raise click.BadParameter(str(ex), param=param)
    if len(src_files) == 0:
        raise click.BadParameter('No source image(s) found.', param=param)
    return src_files


def _raster_file_cb(ctx: click.Context, param: click.Parameter, path_uri: str) -> OpenFile:
    """Click callback to convert a file path / URI to an OpenFile instance in binary read mode."""
    ofile = None
    if path_uri:
        try:
            ofile = fsspec.open(path_uri, 'rb')
        except Exception as ex:
            raise click.BadParameter(str(ex), param=param)
    return ofile


def _text_file_cb(ctx: click.Context, param: click.Parameter, path_uri: str) -> OpenFile:
    """Click callback to convert a file path / URI to an OpenFile instance in text read mode."""
    ofile = None
    if path_uri:
        try:
            ofile = fsspec.open(path_uri, 'rt')
        except Exception as ex:
            raise click.BadParameter(str(ex), param=param)
    return ofile


def _lla_crs_cb(ctx: click.Context, param: click.Parameter, lla_crs: str) -> str:
    """Click callback to validate and parse the LLA CRS."""
    if lla_crs is not None:
        try:
            lla_crs = _read_crs(lla_crs)
        except Exception as ex:
            raise click.BadParameter(f'{str(ex)}', param=param)
        if not lla_crs.is_geographic:
            raise click.BadParameter(f"CRS should be a geographic system.", param=param)
    return lla_crs


def _resolution_cb(
    ctx: click.Context, param: click.Parameter, resolution: tuple
) -> tuple[float, float]:
    """Click callback to validate and parse the resolution."""
    if len(resolution) == 1:
        resolution *= 2
    elif len(resolution) > 2:
        raise click.BadParameter(f'At most two resolution values should be specified.', param=param)
    return resolution


def _dir_cb(ctx: click.Context, param: click.Parameter, uri_path: str) -> OpenFile:
    """Click callback to convert a directory to an fsspec OpenFile, and validate."""
    try:
        # some dirs (e.g. gcs buckets) don't work with a trailing slash on the path, so strip it
        ofile = fsspec.open(uri_path.rstrip('/'))
    except Exception as ex:
        raise click.BadParameter(str(ex), param=param)

    if not ofile.fs.isdir(ofile.path):
        raise click.BadParameter(
            f"'{uri_path}' is not a directory or cannot be accessed.", param=param
        )
    return ofile


def _odm_out_dir_cb(ctx: click.Context, param: click.Parameter, out_dir: str) -> OpenFile:
    """Click callback to create the default, or validate a user-supplied, ODM output directory."""
    if not out_dir:
        ofile = utils.join_ofile(ctx.params['dataset_dir'], 'orthority')
        try:
            # Note this does not raise an exception for (read-only) html protocol.  Not ideal, but
            # this code won't be reached for html as _dir_cb does raise an exception for
            # --dataset-dir.
            ofile.fs.mkdirs(ofile.path, exist_ok=True)
        except Exception as ex:
            raise click.BadParameter(f"Cannot create / access '{out_dir}'. {str(ex)}", param=param)
    else:
        ofile = _dir_cb(ctx, param, out_dir)
    return ofile


def _get_bar_format(units: str = 'files', desc_width: int = 0, units_width: int = None) -> str:
    """Return a ``tqdm`` ``bar_format`` for the given arguments."""
    # pad or truncate desc to desc_width
    lbar = f"{{desc:<{desc_width}.{desc_width}}}:" + "{percentage:3.0f}%|"
    rbar = '|{n_fmt}/{total_fmt} ' + units + ' [{elapsed}<{remaining}]'
    # find a bar width that gives enough space for rhs stats with 3 digit n_fmt / total_fmt and
    # elapsed / remaining as hh:mm:ss
    if not units_width:
        units_width = len(units)
    bar_width = max(shutil.get_terminal_size()[0] - (desc_width + 6 + (10 + units_width + 20)), 10)
    bar = f"{{bar:{bar_width}}}"
    return lbar + bar + rbar


def _ortho(
    src_files: Sequence[OpenFile],
    dem_file: OpenFile,
    cameras: Cameras,
    crs: rio.CRS,
    dem_band: int,
    export_params: bool,
    out_dir: OpenFile,
    overwrite: bool,
    **kwargs,
):
    """
    Orthorectify images given a DEM filename and camera factory.

    Backend function for orthorectification sub-commands.
    """
    # TODO: investigate simplifying progress bars with rich

    if export_params:
        # convert interior / exterior params to oty format files & exit
        logger.info('Writing parameter files...')
        try:
            cameras.write_param(out_dir, overwrite=overwrite)
        except FileExistsError as ex:
            raise click.UsageError(str(ex))
        except CrsMissingError:
            raise click.MissingParameter(param_hint="'-c' / '--crs'", param_type='option')
        return
    elif not dem_file:
        raise click.MissingParameter(param_hint="'-d' / '--dem'", param_type='option')

    # set up progress bar formats
    src_name_lens = [len(Path(src_file.path).name) for src_file in src_files]
    desc_width = min(max(src_name_lens), 40)
    # units_width = max(len('files'), len('blocks'))
    outer_bar_format = _get_bar_format(units='files', desc_width=desc_width, units_width=6)
    inner_bar_format = _get_bar_format(units='blocks', desc_width=desc_width, units_width=6)

    # open & validate dem_file (open it once here so it is not opened repeatedly in
    # orthorectification below)
    try:
        dem_ctx = utils.OpenRaster(dem_file, 'r')
    except FileNotFoundError as ex:
        raise click.BadParameter(str(ex), param_hint="'-d' / '--dem'")

    with dem_ctx as dem_im:
        # validate dem_band
        if dem_band <= 0 or dem_band > dem_im.count:
            dem_name = utils.get_filename(dem_im)
            raise click.BadParameter(
                f"DEM band {dem_band} is invalid for '{dem_name}' with {dem_im.count} band(s).",
                param_hint="'-db' / '--dem-band'",
            )

        for src_file in tqdm(src_files, desc='Total', bar_format=outer_bar_format):
            # set up progress bar args for the current src_file
            src_file_path = Path(src_file.path)
            tqdm_kwargs = dict(desc=src_file_path.name, leave=False, bar_format=inner_bar_format)

            # create camera for src_file
            try:
                camera = cameras.get(src_file)
            except ParamError as ex:
                raise click.UsageError(str(ex))

            # open & validate src_file (open it once here so it is not opened repeatedly)
            try:
                src_ctx = utils.OpenRaster(src_file, 'r')
            except FileNotFoundError as ex:
                raise click.BadParameter(str(ex), param_hint="'SOURCE...'")

            with src_ctx as src_im:
                # finalise and validate world / ortho crs
                crs = crs or src_im.crs
                if not crs:
                    raise click.MissingParameter(param_hint="'-c' / '--crs'", param_type='option')

                # orthorectify
                ortho = Ortho(src_im, dem_im, camera, crs, dem_band=dem_band)
                ortho_ofile = utils.join_ofile(
                    out_dir, f'{src_file_path.stem}_ORTHO.tif', mode='wb'
                )
                try:
                    ortho.process(ortho_ofile, overwrite=overwrite, progress=tqdm_kwargs, **kwargs)
                except FileExistsError as ex:
                    raise click.UsageError(str(ex))


# Define click options that are common to more than one command
src_files_arg = click.argument(
    'src_files',
    nargs=-1,
    metavar='SOURCE...',
    type=click.Path(dir_okay=False),
    callback=_src_files_cb,
)
dem_file_option = click.option(
    '-d',
    '--dem',
    'dem_file',
    type=click.Path(dir_okay=False),
    default=None,
    callback=_raster_file_cb,
    help='Path / URI of a DEM file covering the source image(s).',
)
int_param_file_option = click.option(
    '-ip',
    '--int-param',
    'int_param_file',
    type=click.Path(dir_okay=False),
    required=True,
    default=None,
    callback=_text_file_cb,
    help='Path / URI of an interior parameter file.',
)
ext_param_file_option = click.option(
    '-ep',
    '--ext-param',
    'ext_param_file',
    type=click.Path(dir_okay=False),
    required=True,
    default=None,
    callback=_text_file_cb,
    help='Path / URI of an exterior parameter file.',
)
crs_option = click.option(
    '-c',
    '--crs',
    type=click.STRING,
    default=None,
    show_default='auto',
    callback=_crs_cb,
    help='CRS of ortho image(s) and world coordinates as an EPSG, proj4, or WKT string; path / URI '
    'of a text file containing string; or path / URI of an image with metadata CRS.',
)
lla_crs_option = click.option(
    '-lc',
    '--lla-crs',
    type=click.STRING,
    default='EPSG:4979',
    show_default=True,
    callback=_lla_crs_cb,
    help='CRS of any geographic coordinate exterior parameters as an EPSG, proj4, or WKT string; '
    'path / URI of a text file containing string; or path / URI of an image with metadata CRS.',
)
radians_option = click.option(
    '-rd/-dg',
    '--radians/--degrees',
    type=click.BOOL,
    default=False,
    show_default=True,
    help='Orientation angle units. Only used for ``--ext-param`` in CSV format.',
)
resolution_option = click.option(
    '-r',
    '--res',
    'resolution',
    type=click.FLOAT,
    default=None,
    show_default='ground sampling distance',
    multiple=True,
    callback=_resolution_cb,
    help='Ortho image resolution in units of the ``--crs``.  Can be used twice for non-square '
    'pixels: ``--res PIXEL_WIDTH --res PIXEL_HEIGHT``.',
)
dem_band_option = click.option(
    '-db',
    '--dem-band',
    type=click.INT,
    nargs=1,
    default=Ortho._default_config['dem_band'],
    show_default=True,
    help='Index of the DEM band to use (1 based).',
)
interp_option = click.option(
    '-i',
    '--interp',
    type=click.Choice(Interp, case_sensitive=False),
    default=Ortho._default_config['interp'],
    show_default=True,
    help=f'Interpolation method for remapping source to ortho image.',
)
dem_interp_option = click.option(
    '-di',
    '--dem-interp',
    type=click.Choice(Interp, case_sensitive=False),
    default=Ortho._default_config['dem_interp'],
    show_default=True,
    help=f'Interpolation method for DEM reprojection.',
)
per_band_option = click.option(
    '-pb/-npb',
    '--per-band/--no-per-band',
    type=click.BOOL,
    default=Ortho._default_config['per_band'],
    show_default=True,
    help='Orthorectify band-by-band (``--per-band``) or all bands at once (``--no-per-band``). '
    '``--no-per-band`` is faster but uses more memory.',
)
full_remap_option = click.option(
    '-fr/-nfr',
    '--full-remap/--no-full-remap',
    type=click.BOOL,
    default=FrameCamera._default_distort,
    show_default=True,
    help='Orthorectify the source image with full camera model (``--full-remap``), '
    'or an undistorted source image with pinhole camera model (``--no-full-remap``).  '
    '``--no-full-remap`` is faster but can reduce ortho image quality.',
)
alpha_option = click.option(
    '-a',
    '--alpha',
    type=click.FloatRange(0, 1),
    nargs=1,
    default=1,
    show_default=True,
    help='Scaling of the ``--no-full-remap`` undistorted image: ``0`` includes the largest '
    'source image portion that allows all undistorted pixels to be valid.  ``1`` includes all '
    'source pixels in the undistorted image.',
)
write_mask_option = click.option(
    '-wm/-nwm',
    '--write-mask/--no-write-mask',
    type=click.BOOL,
    default=Ortho._default_config['write_mask'],
    show_default='true for jpeg compression.',
    help='Mask valid pixels with an internal mask (``--write-mask``), or with a nodata value '
    'based on ``--dtype`` (``--no-write-mask``). An internal mask helps remove nodata noise '
    'caused by lossy compression.',
)
dtype_option = click.option(
    '-dt',
    '--dtype',
    type=click.Choice(list(Ortho._nodata_vals.keys()), case_sensitive=False),
    default=Ortho._default_config['dtype'],
    show_default='source image data type.',
    help=f'Ortho image data type.',
)
compress_option = click.option(
    '-c',
    '--compress',
    type=click.Choice(Compress, case_sensitive=False),
    default=Ortho._default_config['compress'],
    show_default="jpeg for uint8 --dtype, deflate otherwise",
    help=f'Ortho image compression.',
)
build_ovw_option = click.option(
    '-bo/-nbo',
    '--build-ovw/--no-build-ovw',
    type=click.BOOL,
    default=True,
    show_default=True,
    help='Build overviews for the ortho image(s).',
)
export_params_option = click.option(
    '-ep',
    '--export-params',
    is_flag=True,
    type=click.BOOL,
    default=False,
    show_default=True,
    help='Export camera parameters to orthority format file(s), and exit.',
)
out_dir_option = click.option(
    '-od',
    '--out-dir',
    type=click.Path(file_okay=False),
    default=str(Path.cwd()),
    show_default='current working',
    callback=_dir_cb,
    help='Path / URI of the output file directory.',
)
overwrite_option = click.option(
    '-o',
    '--overwrite',
    is_flag=True,
    type=click.BOOL,
    default=False,
    show_default=True,
    help='Overwrite existing output(s).',
)


@click.group()
@click.option('--verbose', '-v', count=True, help='Increase verbosity.')
@click.option('--quiet', '-q', count=True, help='Decrease verbosity.')
@click.version_option(version=__version__, message='%(version)s')
@click.pass_context
def cli(ctx: click.Context, verbose, quiet) -> None:
    """Orthorectification toolkit."""
    # configure logging
    verbosity = verbose - quiet
    _configure_logging(verbosity)

    # enter context managers for sub-command raster operations
    env = rio.Env(GDAL_NUM_THREADS='ALL_CPUS', GTIFF_FORCE_RGBA=False, GDAL_TIFF_INTERNAL_MASK=True)
    ctx.with_resource(utils.suppress_no_georef())
    ctx.with_resource(env)

    # redirect logs through tqdm.write, so they do not interfere with progress bars
    ctx.with_resource(logging_redirect_tqdm([logging.getLogger(__package__)], tqdm_class=tqdm))


@cli.command(
    cls=RstCommand,
    short_help='Orthorectify with interior and exterior parameter files.',
    epilog='See https://orthority.readthedocs.io/ for more detail on file formats and usage.',
)
@src_files_arg
@dem_file_option
@int_param_file_option
@ext_param_file_option
@crs_option
@resolution_option
@dem_band_option
@interp_option
@dem_interp_option
@per_band_option
@full_remap_option
@alpha_option
@lla_crs_option
@radians_option
@write_mask_option
@dtype_option
@compress_option
@build_ovw_option
@export_params_option
@out_dir_option
@overwrite_option
def frame(
    int_param_file: OpenFile,
    ext_param_file: OpenFile,
    crs: rio.CRS,
    full_remap: bool,
    alpha: float,
    lla_crs: rio.CRS,
    radians: bool,
    **kwargs,
) -> None:
    """
    Orthorectify SOURCE images with frame camera model(s) defined by interior and exterior
    parameter files.

    SOURCE images can be specified with paths, URIs or path / URI wildcard patterns.

    Interior parameters are supported in Orthority :file:`.yaml`, OpenDroneMap
    :file:`cameras.json`, and OpenSfM :file:`reconstruction.json` formats.  Exterior parameters
    are supported in Orthority :file:`.geojson`, CSV, and OpenSfM :file:`reconstruction.json`
    formats.  Parameter file extensions are used to distinguish their format.

    The :option:`--int-param <oty-frame --int-param>` and :option:`--ext-param <oty-frame
    --ext-param>` options are required.  The :option:`--dem <oty-frame --dem>` option is
    required, except when exporting camera parameters with :option:`--export-params <oty-frame
    --export-params>`.  Depending on the input file formats, :option:`--crs <oty-frame --crs>`
    may also be required::

        oty frame --dem dem.tif --int-param int_param.yaml --ext-param ext_param.csv --crs EPSG:32651 source*.tif

    Camera parameters can be converted into Orthority format files with :option:`--export-params
    <oty-frame --export-params>`::

        oty frame --int-param reconstruction.json --ext-param reconstruction.json --export-params

    Ortho images and parameter files are placed in the current working directory by default. This
    can be changed with :option:`--out-dir <oty-frame --out-dir>`.
    """
    # create camera factory
    try:
        cameras = FrameCameras(
            int_param_file,
            ext_param_file,
            io_kwargs=dict(crs=crs, lla_crs=lla_crs, radians=radians),
            cam_kwargs=dict(distort=full_remap, alpha=alpha),
        )
    except (FileNotFoundError, ParamError) as ex:
        raise click.UsageError(str(ex))
    except CrsMissingError:
        raise click.MissingParameter(param_hint="'-c' / '--crs'", param_type='option')
    except CrsError as ex:
        raise click.BadParameter(str(ex), param_hint="'-c' / '--crs'")

    # get factory CRS if not set already
    crs = crs or cameras.crs

    # orthorectify
    _ortho(cameras=cameras, crs=crs, **kwargs)


@cli.command(
    cls=RstCommand,
    short_help='Orthorectify with image EXIF / XMP tags.',
    epilog='See https://orthority.readthedocs.io/ for more detail.',
)
@src_files_arg
@dem_file_option
@crs_option
@resolution_option
@dem_band_option
@interp_option
@dem_interp_option
@per_band_option
@full_remap_option
@alpha_option
@lla_crs_option
@write_mask_option
@dtype_option
@compress_option
@build_ovw_option
@export_params_option
@out_dir_option
@overwrite_option
def exif(
    src_files: Sequence[OpenFile],
    crs: rio.CRS,
    full_remap: bool,
    alpha: float,
    lla_crs: rio.CRS,
    **kwargs,
) -> None:
    """
    Orthorectify SOURCE images with frame camera model(s) defined by image EXIF / XMP tags.

    SOURCE images can be specified with paths, URIs or path / URI wildcard patterns.

    SOURCE image tags should include DewarpData, focal length & sensor size or 35mm equivalent
    focal length; camera position and camera roll, pitch & yaw.  DewarpData is converted to a
    Brown model if it is present, otherwise a pinhole model is used.  Pinhole approximation and
    tag value accuracy affect ortho image accuracy.

    The :option:`--dem <oty-exif --dem>` option is required, except when exporting camera
    parameters with :option:`--export-params <oty-exif --export-params>`.  If :option:`--crs
    <oty-exif --crs>` is not supplied, a UTM world / ortho CRS is auto-determined from the camera
    positions::

        oty exif --dem dem.tif source*.tif

    Camera parameters can be converted into Orthority format files with :option:`--export-params
    <oty-exif --export-params>`::

        oty exif ---export-params source*.tif

    Ortho images and parameter files are placed in the current working directory by default.
    This can be changed with :option:`--out-dir <oty-exif --out-dir>`.
    """
    # set up progress bar args
    desc = 'Reading parameters'
    bar_format = _get_bar_format(units='files', desc_width=len(desc))
    tqdm_kwargs = dict(desc=desc, bar_format=bar_format, leave=False)

    # create camera factory
    try:
        cameras = FrameCameras.from_images(
            src_files,
            io_kwargs=dict(crs=crs, lla_crs=lla_crs, progress=tqdm_kwargs),
            cam_kwargs=dict(distort=full_remap, alpha=alpha),
        )
    except (FileNotFoundError, ParamError) as ex:
        raise click.BadParameter(str(ex), param_hint="'SOURCE...'")
    except CrsError as ex:
        raise click.BadParameter(str(ex), param_hint="'-c' / '--crs'")

    # get auto UTM CRS, if CRS not set already
    crs = crs or cameras.crs

    # orthorectify
    _ortho(src_files=src_files, cameras=cameras, crs=crs, **kwargs)


@cli.command(
    cls=RstCommand,
    short_help='Orthorectify with OpenDroneMap outputs.',
    epilog='See https://orthority.readthedocs.io/ for more detail.',
)
@click.option(
    '-dd',
    '--dataset-dir',
    type=click.Path(file_okay=False),
    required=True,
    default=None,
    callback=_dir_cb,
    help='Path / URI of the ODM dataset to process.',
)
@crs_option
@resolution_option
@interp_option
@dem_interp_option
@per_band_option
@full_remap_option
@alpha_option
@write_mask_option
@dtype_option
@compress_option
@build_ovw_option
@export_params_option
@click.option(
    '-od',
    '--out-dir',
    type=click.Path(file_okay=False),
    default=None,
    show_default='<dataset-dir>/orthority',
    callback=_odm_out_dir_cb,
    help='Path / URI of the output file directory.',
)
@overwrite_option
def odm(
    dataset_dir: OpenFile,
    crs: rio.CRS,
    resolution: tuple[float, float],
    full_remap: bool,
    alpha: float,
    out_dir: OpenFile,
    **kwargs,
) -> None:
    """
    Orthorectify images in a processed OpenDroneMap dataset that includes a DSM.

    The images, DSM and camera models are read from the dataset. If :option:`--crs <oty-odm
    --crs>` is not supplied (recommended), the world / ortho CRS is also read from the dataset.
    :option:`--dataset-dir <oty-odm --dataset-dir>` is the only required option::

        oty odm --dataset-dir dataset

    Camera parameters can be converted into Orthority format files with :option:`--export-params
    <oty-odm --export-params>`::

        oty odm --dataset-dir dataset --export-params

    Ortho images and parameter files are placed in the :file:`{dataset}/orthority` subdirectory
    by default.  This can be changed with :option:`--out-dir <oty-odm --out-dir>`.
    """
    # find source images
    src_exts = ['.jpg', '.jpeg', '.tif', '.tiff']
    images_ofile = utils.join_ofile(dataset_dir, 'images/')
    src_files = images_ofile.fs.find(images_ofile.path)  # use existing fs rather than open_files
    src_files = [sf for sf in src_files if sf[sf.rfind('.') :].lower() in src_exts]
    src_files = [OpenFile(images_ofile.fs, sf, 'rb') for sf in src_files]
    if len(src_files) == 0:
        raise click.BadParameter(
            f"No images found in '<dataset-dir>/images/'.", param_hint="'-dd' / '--dataset-dir'"
        )

    # read CRS from DSM
    dem_ofile = utils.join_ofile(dataset_dir, 'odm_dem/dsm.tif', mode='rb')
    try:
        dem_ctx = utils.OpenRaster(dem_ofile, 'r')
    except FileNotFoundError as ex:
        raise click.BadParameter(
            f"No DSM found in '<dataset-dir>/odm_dem/'. {str(ex)}",
            param_hint="'-dd' / '--dataset-dir'",
        )
    with dem_ctx as dem_im:
        crs = crs or dem_im.crs

    # create camera factory
    rec_ofile = utils.join_ofile(dataset_dir, 'opensfm/reconstruction.json', mode='rt')
    try:
        cameras = FrameCameras(
            int_param=rec_ofile,
            ext_param=rec_ofile,
            io_kwargs=dict(crs=crs),
            cam_kwargs=dict(distort=full_remap, alpha=alpha),
        )
    except FileNotFoundError as ex:
        raise click.BadParameter(
            f"No 'reconstruction.json' file found in '<dataset-dir>/opensfm/'. {str(ex)}",
            param_hint="'-dd' / '--dataset-dir'",
        )
    except ParamError as ex:
        raise click.UsageError(str(ex))

    # orthorectify
    _ortho(
        src_files=tuple(src_files),
        dem_file=dem_ofile,
        cameras=cameras,
        crs=crs,
        resolution=resolution,
        dem_band=1,
        out_dir=out_dir,
        **kwargs,
    )


@cli.command(
    # make these short help messages consistent between frame / rpc models
    cls=RstCommand,
    short_help='Orthorectify with RPC camera model(s).',
    epilog='See https://orthority.readthedocs.io/ for more detail.',
)
@src_files_arg
@dem_file_option
@click.option(
    '-rp',
    '--rpc-param',
    'rpc_param_file',
    type=click.Path(dir_okay=False),
    default=None,
    callback=_text_file_cb,
    help='Path / URI of an Orthority RPC parameter file.',
)
@crs_option
@resolution_option
@dem_band_option
@interp_option
@dem_interp_option
@per_band_option
@write_mask_option
@dtype_option
@compress_option
@build_ovw_option
@export_params_option
@out_dir_option
@overwrite_option
def rpc(
    src_files: Sequence[OpenFile],
    rpc_param_file: OpenFile,
    crs: rio.CRS,
    **kwargs,
) -> None:
    """
    Orthorectify SOURCE images with RPC camera model(s).

    SOURCE images can be specified with paths, URIs or path / URI wildcard patterns.

    Camera parameters are read from SOURCE image tags / sidecar files, or from
    :option:`--rpc-param <oty-rpc --rpc-param>` if provided.

    The :option:`--dem <oty-rpc --dem>` option is required, except when exporting camera
    parameters with :option:`--export-params <oty-rpc --export-params>`.  If :option:`--crs
    <oty-rpc --crs>` is not supplied, a WGS84 geographic world / ortho CRS is used::

        oty rpc --dem dem.tif source*.tif

    Camera parameters can be converted to an Orthority format file with :option:`--export-params
    <oty-rpc --export-params>`::

        oty rpc ---export-params source*.tif

    Ortho images and parameter files are placed in the current working directory by default.
    This can be changed with :option:`--out-dir <oty-rpc --out-dir>`.
    """
    # set CRS to the RPC camera default (WGS84) if no CRS supplied, otherwise pass user CRS in
    # cam_kwargs
    if not crs:
        crs = rio.CRS.from_epsg(4979)
        cam_kwargs = {}
    else:
        cam_kwargs = dict(crs=crs)

    if rpc_param_file:
        # create camera factory from parameter file
        try:
            cameras = RpcCameras(rpc_param_file, cam_kwargs=cam_kwargs)
        except (FileNotFoundError, ParamError) as ex:
            raise click.BadParameter(str(ex), param_hint="'-rp' / '--rpc-param'")
    else:
        # set up progress bar args
        desc = 'Reading parameters'
        bar_format = _get_bar_format(units='files', desc_width=len(desc))
        tqdm_kwargs = dict(desc=desc, bar_format=bar_format, leave=False)

        # create camera factory from image tags / sidecar file(s)
        try:
            cameras = RpcCameras.from_images(
                src_files, io_kwargs=dict(progress=tqdm_kwargs), cam_kwargs=cam_kwargs
            )
        except (FileNotFoundError, ParamError) as ex:
            raise click.BadParameter(str(ex), param_hint="'SOURCE...'")

    # orthorectify
    _ortho(
        src_files=src_files,
        cameras=cameras,
        crs=crs,
        **kwargs,
    )


def _simple_ortho(
    src_im_file,
    dem_file,
    pos_ori_file,
    ortho_dir=None,
    read_conf=None,
    write_conf=None,
    verbosity=2,
):
    """
    Legacy orthorectification (deprecated).

    Parameters
    ----------
    src_im_file: str, pathlib.Path
        Source image file(s).
    dem_file: str, pathlib.Path
        DEM file covering source image file(s).
    pos_ori_file: str, pathlib.Path
        Position and orientation file for source image file(s).
    ortho_dir: str, pathlib.Path, optional
        Output directory.
    read_conf: str, pathlib.Path, optional
        Read configuration from this file.
    write_conf: str, pathlib.Path, optional
        Write configuration to this file and exit.
    verbosity: int
        Logging verbosity 1=DEBUG, 2=INFO, 3=WARNING, 4=ERROR (default: 2)
    """

    def check_args(src_im_file, dem_file, pos_ori_file, ortho_dir=None):
        """Check arguments for errors."""
        # check files exist
        for src_im_file_spec in src_im_file:
            src_im_file_path = Path(src_im_file_spec)
            if len(list(src_im_file_path.parent.glob(src_im_file_path.name))) == 0:
                raise Exception(f'Could not find any source image(s) matching {src_im_file_spec}')

        if not Path(dem_file).exists():
            raise Exception(f'DEM file {dem_file} does not exist')

        if not Path(pos_ori_file).exists():
            raise Exception(f'Camera position and orientation file {pos_ori_file} does not exist')

        # check and create ortho_dir if necessary
        if ortho_dir is not None:
            ortho_dir = Path(ortho_dir)
            if not ortho_dir.exists():
                logger.warning(f'Creating ortho directory {ortho_dir}')
                ortho_dir.mkdir()

    try:
        # set logging level
        if verbosity is not None:
            _configure_logging(2 - verbosity)

        logger.warning(
            "This command is deprecated and will be removed in a future version.  Please switch to"
            " 'oty' and its sub-commands.\n"
        )

        # read configuration
        if read_conf is None:
            config_filename = root_path.joinpath('config.yaml')
        else:
            config_filename = Path(read_conf)

        if not config_filename.exists():
            raise Exception(f'Config file {config_filename} does not exist')

        with open(config_filename, 'r') as f:
            config = yaml.safe_load(f)

        # write configuration if requested and exit
        if write_conf is not None:
            out_config_filename = Path(write_conf)
            with open(out_config_filename, 'w') as f:
                yaml.dump(config, stream=f)
            logger.info(f'Wrote config to {out_config_filename}')
            return

        # prepare ortho config
        ortho_config = config.get('ortho', {})
        crs = ortho_config.pop('crs', None)
        dem_band = ortho_config.pop('dem_band', Ortho._default_config['dem_band'])
        for key in ['driver', 'tile_size', 'nodata', 'interleave', 'photometric']:
            if key in ortho_config:
                ortho_config.pop(key)
                logger.warning(f"The '{key}' option is deprecated.")
        for key in ['interp', 'dem_interp']:
            if ortho_config.get(key, None) == 'cubic_spline':
                logger.warning(
                    f"'cubic_spline' interpolation is deprecated, using '{key}'='cubic'."
                )
                ortho_config[key] = 'cubic'
        distort = ortho_config.pop('full_remap', True)

        # prepare camera config
        camera = None
        camera_config = config['camera']
        camera_type = CameraType(camera_config.get('type', 'pinhole'))
        camera_config = {k: v for k, v in camera_config.items() if k not in ['name', 'type']}

        # checks paths etc
        check_args(src_im_file, dem_file, pos_ori_file, ortho_dir=ortho_dir)

        # read camera position and rotation
        with open(pos_ori_file, 'r', newline='') as f:
            reader = csv.DictReader(
                f,
                delimiter=' ',
                fieldnames=['file', 'x', 'y', 'z', 'omega', 'phi', 'kappa'],
            )
            cam_pos_orid = {
                row['file']: {k: float(row[k]) for k in reader.fieldnames[1:]} for row in reader
            }

        # loop through image file(s) or wildcard(s), or combinations thereof
        for src_im_file_spec in src_im_file:
            src_im_file_path = Path(src_im_file_spec)
            for src_filename in src_im_file_path.parent.glob(src_im_file_path.name):
                if src_filename.stem not in cam_pos_orid:
                    raise Exception(f'Could not find {src_filename.stem} in {pos_ori_file}')

                im_pos_ori = cam_pos_orid[src_filename.stem]
                opk = np.radians((im_pos_ori['omega'], im_pos_ori['phi'], im_pos_ori['kappa']))
                xyz = np.array((im_pos_ori['x'], im_pos_ori['y'], im_pos_ori['z']))

                # set ortho filename
                ortho_dir = src_filename.parent if not ortho_dir else ortho_dir
                ortho_filename = Path(ortho_dir).joinpath(src_filename.stem + '_ORTHO.tif')

                # Get src size
                with utils.suppress_no_georef(), rio.open(src_filename) as src_im:
                    im_size = np.float64([src_im.width, src_im.height])

                if not camera:
                    # create a new camera
                    camera = create_camera(
                        camera_type, **camera_config, xyz=xyz, opk=opk, distort=distort
                    )
                else:
                    # update existing camera
                    camera.update(xyz, opk)

                if np.any(im_size != camera._im_size):
                    logger.warning(
                        f'{src_filename.name} size ({im_size}) does not match configuration.'
                    )

                # create Ortho and orthorectify
                logger.info(f'Orthorectifying {src_filename.name}:')
                ortho_im = Ortho(src_filename, dem_file, camera, crs=crs, dem_band=dem_band)
                ortho_im.process(ortho_filename, **ortho_config)

    except Exception as ex:
        logger.error('Exception: ' + str(ex))
        raise ex


def _get_simple_ortho_parser():
    """Return an argparse parser for the legacy ``simple-ortho`` CLI."""
    parser = argparse.ArgumentParser(
        description='Orthorectify an image with known DEM and camera model.'
    )
    parser.add_argument(
        "src_im_file",
        help="path(s) and or wildcard(s) specifying the source image file(s)",
        type=str,
        metavar='src_im_file',
        nargs='+',
    )
    parser.add_argument("dem_file", help="path to the DEM file", type=str)
    parser.add_argument(
        "pos_ori_file", help="path to the camera position and orientation file", type=str
    )
    parser.add_argument(
        "-od",
        "--ortho-dir",
        help="write ortho image(s) to this directory (default: write ortho image(s) to source directory)",
        type=str,
    )
    parser.add_argument(
        "-rc",
        "--read-conf",
        help="read custom config from this path (default: use config.yaml in simple-ortho root)",
        type=str,
    )
    parser.add_argument(
        "-wc", "--write-conf", help="write default config to this path and exit", type=str
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        choices=[1, 2, 3, 4],
        help="logging level: 1=DEBUG, 2=INFO, 3=WARNING, 4=ERROR (default: 2)",
        type=int,
    )
    return parser


def simple_ortho(argv=None):
    """Entry point to legacy ``simple-ortho`` CLI."""
    args = _get_simple_ortho_parser().parse_args(argv)
    _simple_ortho(**vars(args))


if __name__ == '__main__':
    cli()

##



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import shutil
import tracemalloc
from pathlib import Path

import click
import fsspec
import numpy as np
import pytest
import rasterio as rio
import yaml
from click.testing import CliRunner

from orthority.cli import _ortho, cli, simple_ortho
from orthority.factory import FrameCameras

logger = logging.getLogger(__name__)


@pytest.fixture(scope='session')
def frame_legacy_ngi_cli_str(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_legacy_config_file: Path,
    ngi_legacy_csv_file: Path,
) -> str:
    """``oty frame`` CLI string to orthorectify an NGI image with legacy format interior and
    exterior params.
    """
    return (
        f'frame --dem {ngi_dem_file} --int-param {ngi_legacy_config_file} '
        f'--ext-param {ngi_legacy_csv_file} {ngi_image_file}'
    )


def test_oty_verbosity(
    frame_legacy_ngi_cli_str: str, ngi_image_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty -v frame`` generates debug logs."""
    cli_str = f'-v {frame_legacy_ngi_cli_str} --res 50 --out-dir {tmp_path}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    assert 'DEBUG:' in result.stdout


def test_frame_help(runner: CliRunner):
    """Test ``oty frame --help``."""
    result = runner.invoke(cli, 'frame --help'.split())
    assert result.exit_code == 0, result.stdout
    assert len(result.stdout) > 0


def test_frame_src_file_wildcard(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_int_param_file: Path,
    ngi_oty_ext_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` with a SOURCE image wildcard (that is not expanded by the console)."""
    src_wildcard = ngi_image_file.parent.joinpath('*RGB.tif')
    cli_str = (
        f'frame --dem {ngi_dem_file} --int-param {ngi_oty_int_param_file} '
        f'--ext-param {ngi_oty_ext_param_file} --res 30 --out-dir {tmp_path}'
        f' {src_wildcard.as_posix()}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == len([*src_wildcard.parent.glob(src_wildcard.name)])


def test_frame_src_file_not_found_error(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_int_param_file: Path,
    ngi_oty_ext_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` with a non-existing SOURCE image raises an error ."""
    # use a source path whose file name is in the exterior params, but the path doesn't exist
    src_file = f'unknown/{ngi_image_file.name}'
    cli_str = (
        f'frame --dem {ngi_dem_file} --int-param {ngi_oty_int_param_file} '
        f'--ext-param {ngi_oty_ext_param_file} --out-dir {tmp_path} {src_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        'SOURCE' in result.stdout and 'No such file' in result.stdout and src_file in result.stdout
    )


def test_frame_dem_missing_error(
    ngi_image_file: Path,
    ngi_oty_int_param_file: Path,
    ngi_oty_ext_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` without ``--dem`` raises an error."""
    cli_str = (
        f'frame --int-param {ngi_oty_int_param_file} --ext-param {ngi_oty_ext_param_file} '
        f'--out-dir {tmp_path} {ngi_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--dem' in result.stdout and 'missing' in result.stdout.lower()


def test_frame_dem_not_found_error(
    ngi_image_file: Path,
    ngi_oty_int_param_file: Path,
    ngi_oty_ext_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` with a non-existing ``--dem`` raises an error ."""
    dem_file = 'unknown.tif'
    cli_str = (
        f'frame --dem {dem_file} --int-param {ngi_oty_int_param_file} '
        f'--ext-param {ngi_oty_ext_param_file} --out-dir {tmp_path} {ngi_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        '--dem' in result.stdout and 'No such file' in result.stdout and dem_file in result.stdout
    )


def test_frame_int_param_missing_error(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_ext_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` without ``--int-param`` raises an error ."""
    cli_str = (
        f'frame --dem {ngi_dem_file} --ext-param {ngi_oty_ext_param_file} --out-dir {tmp_path} '
        f'{ngi_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--int-param' in result.stdout and 'missing' in result.stdout.lower()


def test_frame_int_param_not_found_error(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_ext_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` with a non-existing ``--int-param`` raises an error ."""
    int_param_file = 'unknown.yaml'
    cli_str = (
        f'frame --dem {ngi_dem_file} --int-param {int_param_file} '
        f'--ext-param {ngi_oty_ext_param_file} --out-dir {tmp_path} {ngi_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert 'No such file' in result.stdout and int_param_file in result.stdout


def test_frame_int_param_ext_error(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_ext_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` with an unrecognised ``--int-param`` extension raises an error."""
    cli_str = (
        f'frame --dem {ngi_dem_file} --int-param {ngi_oty_ext_param_file} '
        f'--ext-param {ngi_oty_ext_param_file} --out-dir {tmp_path} {ngi_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert 'not supported' in result.stdout.lower()


def test_frame_ext_param_missing_error(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_int_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` without ``--ext-param`` raises an error ."""
    cli_str = (
        f'frame --dem {ngi_dem_file} --int-param {ngi_oty_int_param_file} --out-dir {tmp_path} '
        f'{ngi_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--ext-param' in result.stdout and 'missing' in result.stdout.lower()


def test_frame_ext_param_not_found_error(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_int_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` with a non-existing ``--ext-param`` raises an error ."""
    ext_param_file = 'unknown.geojson'
    cli_str = (
        f'frame --dem {ngi_dem_file} --int-param {ngi_oty_int_param_file} '
        f'--ext-param {ext_param_file} --out-dir {tmp_path} {ngi_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert 'No such file' in result.stdout and ext_param_file in result.stdout


def test_frame_ext_param_ext_error(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_int_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` with an unrecognised ``--ext-param`` extension raises an error."""
    cli_str = (
        f'frame --dem {ngi_dem_file} --int-param {ngi_oty_int_param_file} '
        f'--ext-param {ngi_oty_int_param_file} --out-dir {tmp_path} {ngi_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert 'not supported' in result.stdout.lower()


def test_frame_crs_src(
    frame_legacy_ngi_cli_str: str, ngi_image_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty frame`` reads the world / ortho CRS from a projected source image."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ngi_image_file, 'r') as src_im, rio.open(ortho_files[0], 'r') as ortho_im:
        assert src_im.crs == ortho_im.crs


def test_frame_crs_auto(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    odm_lla_rpy_csv_file: Path,
    odm_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` auto-determines the CRS for LLA-RPY CSV format exterior parameters."""
    cli_str = (
        f'frame --dem {odm_dem_file} --int-param {odm_reconstruction_file} '
        f'--ext-param {odm_lla_rpy_csv_file} --out-dir {tmp_path} --res 5 {odm_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.crs == rio.CRS.from_string(odm_crs)


def test_frame_crs_cli(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    odm_xyz_opk_csv_file: Path,
    odm_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` uses a CRS specified with ``--crs``."""
    # use odm_xyz_opk_csv_file exterior params so there is no auto-crs
    cli_str = (
        f'frame --dem {odm_dem_file} --int-param {odm_reconstruction_file} '
        f'--ext-param {odm_xyz_opk_csv_file} --out-dir {tmp_path} --crs {odm_crs} --res 5 '
        f'{odm_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.crs == rio.CRS.from_string(odm_crs)


def test_frame_crs_file(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    odm_xyz_opk_csv_file: Path,
    odm_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` uses a text file CRS specified with ``--crs``."""
    # use odm_xyz_opk_csv_file exterior params so there is no auto-crs
    crs_file = tmp_path.joinpath('test_crs.txt')
    crs_file.write_text(odm_crs)
    # use file:// prefix to test differentiation with CRS string e.g. 'EPSG:...'
    cli_str = (
        f'frame --dem {odm_dem_file} --int-param {odm_reconstruction_file} '
        f'--ext-param {odm_xyz_opk_csv_file} --out-dir {tmp_path} --crs {crs_file} --res 5 '
        f'{odm_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.crs == rio.CRS.from_string(odm_crs)


def test_frame_crs_prj(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    odm_xyz_opk_csv_file: Path,
    odm_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` reads the CRS in a CSV exterior parameter .prj file."""
    # copy csv file to tmp_path and create .prj file
    csv_file = tmp_path.joinpath(odm_xyz_opk_csv_file.name)
    csv_file.write_text(odm_xyz_opk_csv_file.read_text())
    prj_file = csv_file.with_suffix('.prj')
    prj_file.write_text(odm_crs)

    # create ortho & test
    cli_str = (
        f'frame --dem {odm_dem_file} --int-param {odm_reconstruction_file} --ext-param {csv_file} '
        f'--out-dir {tmp_path} --res 5 {odm_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.crs == rio.CRS.from_string(odm_crs)


def test_frame_crs_missing_error(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    odm_xyz_opk_csv_file: Path,
    odm_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` raises an error when ``--crs`` is needed but not passed."""
    cli_str = (
        f'frame --dem {odm_dem_file} --int-param {odm_reconstruction_file} '
        f'--ext-param {odm_xyz_opk_csv_file} --out-dir {tmp_path} {odm_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--crs' in result.stdout and 'missing' in result.stdout.lower()


def test_frame_crs_geographic_error(
    frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty frame`` raises an error when ``--crs`` is geographic."""
    cli_str = frame_legacy_ngi_cli_str + ' --crs EPSG:4326'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--crs' in result.stdout and 'projected' in result.stdout


def test_frame_crs_invalid_error(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame`` raises an error when ``--crs`` is invalid."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --crs unknown'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--crs' in result.stdout and 'invalid' in result.stdout.lower()


def test_frame_resolution_square(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --res`` with square resolution."""
    resolution = (96.0, 96.0)
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --res {resolution[0]}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.res == resolution


def test_frame_resolution_non_square(
    frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty frame --res`` with non-square resolution."""
    resolution = (48.0, 96.0)
    cli_str = (
        frame_legacy_ngi_cli_str
        + f' --out-dir {tmp_path} --res {resolution[0]} --res {resolution[1]}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.res == resolution


def test_frame_resolution_auto(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame`` generates an ortho without the ``--res`` option."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1


def test_frame_dem_band(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --dem-band`` with valid band."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --res 24 --dem-band 1'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1


def test_frame_dem_band_error(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --dem-band`` raises an error with an out of range band."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --res 24 --dem-band 2'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--dem-band' in result.stdout


def test_frame_interp(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --interp`` by comparing ``nearest`` to ``average`` interpolation orthos."""
    ortho_std = []
    for interp in ['average', 'nearest']:
        # create ortho
        out_dir = tmp_path.joinpath(interp)
        out_dir.mkdir()
        cli_str = (
            frame_legacy_ngi_cli_str
            + f' --out-dir {out_dir} --res 24 --compress deflate --interp {interp}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert result.exit_code == 0, result.stdout
        ortho_files = [*out_dir.glob('*_ORTHO.tif')]
        assert len(ortho_files) == 1

        # read ortho std dev
        with rio.open(ortho_files[0], 'r') as im:
            array = im.read(masked=True)
        ortho_std.append(float(array.std()))

    # compare orthos
    assert ortho_std[0] != pytest.approx(ortho_std[1], abs=0.1)


def test_frame_interp_error(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --interp`` raises an error with an invalid interpolation value."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --interp other'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--interp' in result.stdout and 'invalid' in result.stdout.lower()


def test_frame_dem_interp(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --dem-interp`` by comparing ``nearest`` to ``average`` DEM interpolation
    orthos.
    """
    ortho_std = []
    for dem_interp in ['average', 'nearest']:
        # create ortho
        out_dir = tmp_path.joinpath(dem_interp)
        out_dir.mkdir()
        cli_str = (
            frame_legacy_ngi_cli_str
            + f' --out-dir {out_dir} --res 30 --compress deflate --dem-interp {dem_interp}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert result.exit_code == 0, result.stdout
        ortho_files = [*out_dir.glob('*_ORTHO.tif')]
        assert len(ortho_files) == 1

        # read ortho std dev
        with rio.open(ortho_files[0], 'r') as im:
            array = im.read(masked=True)
        ortho_std.append(float(array.std()))

    # compare orthos
    assert ortho_std[0] != pytest.approx(ortho_std[1], abs=0.01)


def test_frame_dem_interp_error(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --dem-interp`` raises an error with an invalid interpolation value."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --dem-interp other'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--dem-interp' in result.stdout and 'invalid' in result.stdout.lower()


def test_frame_per_band(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --per-band`` by comparing memory usage between ``--per-band`` and ``--no-
    per-band``.
    """
    mem_peaks = []
    tracemalloc.start()
    try:
        for per_band in ['per-band', 'no-per-band']:
            # create ortho
            out_dir = tmp_path.joinpath(per_band)
            out_dir.mkdir()
            cli_str = (
                frame_legacy_ngi_cli_str
                + f' --out-dir {out_dir} --res 30 --compress deflate --{per_band}'
            )
            result = runner.invoke(cli, cli_str.split())
            assert result.exit_code == 0, result.stdout
            ortho_files = [*out_dir.glob('*_ORTHO.tif')]
            assert len(ortho_files) == 1

            _, mem_peak = tracemalloc.get_traced_memory()
            tracemalloc.clear_traces()
            mem_peaks.append(mem_peak)

        assert mem_peaks[1] > mem_peaks[0]
    finally:
        tracemalloc.stop()


def test_frame_full_remap(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame --full-remap`` by comparing ``--full-remap`` and ``--no-full-remap``
    orthos.
    """
    ortho_arrays = []
    for full_remap in ['full-remap', 'no-full-remap']:
        # create ortho
        out_dir = tmp_path.joinpath(full_remap)
        out_dir.mkdir()
        cli_str = (
            f'frame --dem {odm_dem_file} --int-param {odm_reconstruction_file} '
            f'--ext-param {odm_reconstruction_file} --out-dir {out_dir} --res 1 '
            f'--compress deflate --{full_remap} {odm_image_file}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert result.exit_code == 0, result.stdout
        ortho_files = [*out_dir.glob('*_ORTHO.tif')]
        assert len(ortho_files) == 1

        # read ortho
        with rio.open(ortho_files[0], 'r') as im:
            array = im.read(masked=True)
        ortho_arrays.append(array)

    # compare ortho masks
    assert np.any(ortho_arrays[1].mask != ortho_arrays[0].mask)


def test_frame_alpha(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame --alpha`` by comparing ``--alpha 0`` and ``--alpha 1`` orthos."""
    ortho_bounds = []
    for alpha in [0, 1]:
        # create ortho
        out_dir = tmp_path.joinpath(str(alpha))
        out_dir.mkdir()
        cli_str = (
            f'frame --dem {odm_dem_file} --int-param {odm_reconstruction_file} '
            f'--ext-param {odm_reconstruction_file} --out-dir {out_dir} --res 1 '
            f'--compress deflate --no-full-remap --alpha {alpha} {odm_image_file}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert result.exit_code == 0, result.stdout
        ortho_files = [*out_dir.glob('*_ORTHO.tif')]
        assert len(ortho_files) == 1

        # read ortho bounds
        with rio.open(ortho_files[0], 'r') as im:
            ortho_bounds.append(im.bounds)

    # compare ortho bounds
    assert ortho_bounds[0][:2] > ortho_bounds[1][:2] and ortho_bounds[0][-2:] < ortho_bounds[1][-2:]


def test_frame_alpha_error(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --alpha`` raises an error with an invalid alpha value."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --no-full-remap --alpha 2'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--alpha' in result.stdout and 'invalid' in result.stdout.lower()


def test_frame_lla_crs(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    odm_lla_rpy_csv_file: Path,
    odm_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame --lla-crs`` by comparing orthos created with different ``--lla-crs``
    values.
    """
    ortho_bounds = []
    res = 5
    for i, lla_crs in enumerate(['EPSG:4326+4326', 'EPSG:4326+3855']):
        # create ortho
        out_dir = tmp_path.joinpath(str(i))
        out_dir.mkdir()
        cli_str = (
            f'frame --dem {odm_dem_file} --int-param {odm_reconstruction_file} '
            f'--ext-param {odm_lla_rpy_csv_file} --out-dir {out_dir} --res {res} '
            f'--crs {odm_crs}+3855 --lla-crs {lla_crs} {odm_image_file}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert result.exit_code == 0, result.stdout
        ortho_files = [*out_dir.glob('*_ORTHO.tif')]
        assert len(ortho_files) == 1

        # read ortho bounds
        with rio.open(ortho_files[0], 'r') as im:
            ortho_bounds.append(im.bounds)

    # compare ortho bounds
    assert ortho_bounds[1] != pytest.approx(ortho_bounds[0], abs=res)


def test_frame_lla_crs_projected_error(
    frame_legacy_ngi_cli_str: str, ngi_crs: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty frame --lla-crs`` raises an error with a projected CRS value."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --lla-crs {ngi_crs}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--lla-crs' in result.stdout and 'geographic' in result.stdout


def test_frame_lla_crs_invalid_error(
    frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty frame`` raises an error when ``--lla_crs`` is invalid."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --lla-crs unknown'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--lla-crs' in result.stdout and 'invalid' in result.stdout.lower()


def test_frame_radians(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_int_param_file: Path,
    ngi_xyz_opk_csv_file: Path,
    ngi_xyz_opk_radians_csv_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame --radians`` by comparing orthos created with exterior parameter orientation
    angles in radians and degrees.
    """
    ortho_arrays = []
    for radians, ext_param_file in zip(
        ['degrees', 'radians'], [ngi_xyz_opk_csv_file, ngi_xyz_opk_radians_csv_file]
    ):
        # create ortho
        out_dir = tmp_path.joinpath(radians)
        out_dir.mkdir()
        cli_str = (
            f'frame --dem {ngi_dem_file} --int-param {ngi_oty_int_param_file} '
            f'--ext-param {ext_param_file} --out-dir {out_dir} --res 24 '
            f'--{radians} {ngi_image_file}'
        )
        result = runner.invoke(cli, cli_str.split())
        assert result.exit_code == 0, result.stdout
        ortho_files = [*out_dir.glob('*_ORTHO.tif')]
        assert len(ortho_files) == 1

        # read ortho
        with rio.open(ortho_files[0], 'r') as im:
            ortho_arrays.append(im.read(1))

    # compare ortho similarity
    assert ortho_arrays[1] == pytest.approx(ortho_arrays[0], abs=1)


def test_frame_write_mask(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --write-mask`` writes an internal mask to the ortho with
    ``compress=deflate``.
    """
    cli_str = (
        frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --compress deflate --write-mask --res 24'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.profile['compress'] == 'deflate'
        assert all([mf[0] == rio.enums.MaskFlags.per_dataset for mf in im.mask_flag_enums])


def test_frame_dtype(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --dtype`` creates an ortho with the correct dtype."""
    dtype = 'float32'
    cli_str = (
        frame_legacy_ngi_cli_str
        + f' --out-dir {tmp_path} --compress deflate --dtype {dtype} --res 24'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.profile['dtype'] == dtype


def test_frame_dtype_error(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --dtype`` with an invalid dtype raises an error."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --compress deflate --dtype int32'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--dtype' in result.stdout and 'invalid' in result.stdout.lower()


@pytest.mark.parametrize('compress', ['jpeg', 'deflate'])
def test_frame_compress(
    frame_legacy_ngi_cli_str: str, compress: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty frame --compress`` creates an ortho with the correct compression."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --compress {compress} --res 24'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.profile['compress'] == compress


def test_frame_compress_error(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --dtype`` with an invalid dtype raises an error."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --compress other'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--compress' in result.stdout and 'invalid' in result.stdout.lower()


def test_frame_build_ovw(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --build-ovw`` builds overviews."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --res 5 --build-ovw'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert len(im.overviews(1)) > 0


def test_frame_no_build_ovw(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame --no-build-ovw`` does not build overviews."""
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --res 5 --no-build-ovw'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert len(im.overviews(1)) == 0


@pytest.mark.parametrize(
    'int_param_file, ext_param_file',
    [
        ('ngi_oty_int_param_file', 'ngi_oty_ext_param_file'),
        ('ngi_legacy_config_file', 'ngi_xyz_opk_csv_file'),
        ('odm_reconstruction_file', 'odm_reconstruction_file'),
    ],
)
def test_frame_export_params(
    int_param_file: str,
    ext_param_file: str,
    tmp_path: Path,
    runner: CliRunner,
    request: pytest.FixtureRequest,
):
    """Test ``oty frame --export-params`` exports interior & exterior parameters provided in
    different formats (with no ``--dem-file``).
    """
    # NOTE this doubles as a test of reading params in different formats
    int_param_file: Path = request.getfixturevalue(int_param_file)
    ext_param_file: Path = request.getfixturevalue(ext_param_file)
    cli_str = (
        f'frame --int-param {int_param_file} --ext-param {ext_param_file} --out-dir {tmp_path} '
        f'--export-params'
    )

    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    int_param_file = tmp_path.joinpath('int_param.yaml')
    ext_param_file = tmp_path.joinpath('ext_param.geojson')
    assert int_param_file.exists() and ext_param_file.exists()


def test_frame_out_dir_error(frame_legacy_ngi_cli_str: str, tmp_path: Path, runner: CliRunner):
    """Test ``oty frame`` raises an error when --out-dir does not exist."""
    out_dir = 'unknown'
    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {out_dir} --res 24'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        '--out-dir' in result.stdout and 'directory' in result.stdout and out_dir in result.stdout
    )


def test_frame_overwrite(
    frame_legacy_ngi_cli_str: str, ngi_image_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty frame --overwrite`` overwrites an existing ortho."""
    ortho_file = tmp_path.joinpath(ngi_image_file.stem + '_ORTHO.tif')
    ortho_file.touch()

    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --res 24 --overwrite'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1
    assert ortho_file == ortho_files[0]


def test_frame_overwrite_error(
    frame_legacy_ngi_cli_str: str, ngi_image_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty frame`` raises an error when the ortho file already exists."""
    ortho_file = tmp_path.joinpath(ngi_image_file.stem + '_ORTHO.tif')
    ortho_file.touch()

    cli_str = frame_legacy_ngi_cli_str + f' --out-dir {tmp_path} --res 24'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert 'exists' in result.stdout and ortho_file.name in result.stdout


def test_frame_urls(
    ngi_image_url: str,
    ngi_dem_url: str,
    ngi_oty_int_param_url: str,
    ngi_oty_ext_param_url: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty frame`` with source, DEM & parameter file URLs."""
    cli_str = (
        f'frame --dem {ngi_dem_url} --int-param {ngi_oty_int_param_url} '
        f'--ext-param {ngi_oty_ext_param_url} --out-dir {tmp_path} {ngi_image_url}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1


def test_exif_source_error(
    ngi_image_file: Path, ngi_dem_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty exif`` raises an error with a non-EXIF source image."""
    cli_str = f'exif --dem {ngi_dem_file} --out-dir {tmp_path} {ngi_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert 'SOURCE' in result.stdout and 'tags' in result.stdout.lower()


def test_exif_source_not_found_error(
    ngi_image_file: Path, ngi_dem_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty exif`` raises an error with a non-existing source image."""
    src_file = 'unknown.tif'
    cli_str = f'exif --dem {ngi_dem_file} --out-dir {tmp_path} {src_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        'SOURCE' in result.stdout
        and 'No such file' in result.stdout
        and src_file in result.stdout.lower()
    )


def test_exif_crs(
    odm_image_file: Path, odm_dem_file: Path, odm_crs: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty exif --crs`` creates orthos with the correct CRS."""
    crs = odm_crs + '+3855'  # differentiate from auto CRS
    cli_str = f'exif --dem {odm_dem_file} --out-dir {tmp_path} --res 5 --crs {crs} {odm_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as ortho_im:
        assert ortho_im.crs == rio.CRS.from_string(crs)


def test_exif_crs_auto(
    odm_image_file: Path, odm_dem_file: Path, odm_crs: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty exif`` auto determines the CRS."""
    cli_str = f'exif --dem {odm_dem_file} --out-dir {tmp_path} --res 5 {odm_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as ortho_im:
        assert ortho_im.crs == rio.CRS.from_string(odm_crs)


def test_exif_crs_geographic_error(
    odm_image_file: Path, odm_dem_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty exif`` raises an error when ``--crs`` is geographic."""
    cli_str = f'exif --dem {odm_dem_file} --out-dir {tmp_path} --crs EPSG:4326 {odm_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--crs' in result.stdout and 'projected' in result.stdout


def test_exif_crs_invalid_error(
    odm_image_file: Path, odm_dem_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty exif`` raises an error when ``--crs`` is invalid."""
    cli_str = f'exif --dem {odm_dem_file} --out-dir {tmp_path} --crs unknown {odm_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert '--crs' in result.stdout and 'invalid' in result.stdout.lower()


def test_exif_option(
    odm_image_file: Path, odm_dem_file: Path, odm_crs: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty exif`` passes through a ``--res`` option."""
    res = 5
    cli_str = f'exif --dem {odm_dem_file} --out-dir {tmp_path} --res {res} {odm_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as ortho_im:
        assert ortho_im.res == (res, res)


def test_exif_lla_crs(
    odm_image_file: Path,
    odm_dem_file: Path,
    odm_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty exif --lla-crs`` by comparing orthos created with different ``--lla-crs``
    values.
    """
    ortho_bounds = []
    res = 5
    for i, lla_crs in enumerate(['EPSG:4979', 'EPSG:4326+3855']):
        # create ortho
        out_dir = tmp_path.joinpath(str(i))
        out_dir.mkdir()
        cli_str = (
            f'exif --dem {odm_dem_file} --out-dir {out_dir} --res 5 --crs {odm_crs}+3855 '
            f'--lla-crs {lla_crs} {odm_image_file}'
        )

        result = runner.invoke(cli, cli_str.split())
        assert result.exit_code == 0, result.stdout
        ortho_files = [*out_dir.glob('*_ORTHO.tif')]
        assert len(ortho_files) == 1

        # read ortho bounds
        with rio.open(ortho_files[0], 'r') as im:
            ortho_bounds.append(im.bounds)

    # compare ortho bounds
    assert ortho_bounds[1] != pytest.approx(ortho_bounds[0], abs=res / 2)


def test_odm_dataset_dir(
    odm_dataset_dir: Path, odm_image_files: tuple[Path, ...], tmp_path: Path, runner: CliRunner
):
    """Test ``oty odm`` creates orthos in ``<--dataset-dir>/orthority`` sub-folder."""
    shutil.copytree(odm_dataset_dir, tmp_path, dirs_exist_ok=True)  # copy test data to tmp_path
    cli_str = f'odm --dataset-dir {tmp_path} --res 5 --overwrite'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.joinpath('orthority').glob('*_ORTHO.tif')]
    src_files = [*tmp_path.joinpath('images').glob('*.tif')]
    assert len(ortho_files) == len(src_files)


def test_odm_out_dir(
    odm_dataset_dir: Path, odm_image_files: tuple[Path, ...], tmp_path: Path, runner: CliRunner
):
    """Test ``oty odm --out-dir`` creates orthos in the ``--out-dir`` folder."""
    cli_str = f'odm --dataset-dir {odm_dataset_dir} --res 5 --out-dir {tmp_path}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    src_files = [*odm_dataset_dir.joinpath('images').glob('*.tif')]
    assert len(ortho_files) == len(src_files)


def test_odm_option(
    odm_dataset_dir: Path, odm_dem_file: Path, odm_crs: str, tmp_path: Path, runner: CliRunner
):
    """Test ``oty odm`` passes through a ``--res`` option."""
    res = 5
    cli_str = f'odm --dataset-dir {odm_dataset_dir} --res {res} --out-dir {tmp_path}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]

    with rio.open(odm_dem_file, 'r') as dem_im:
        dem_crs = dem_im.crs
    for ortho_file in ortho_files:
        with rio.open(ortho_file, 'r') as ortho_im:
            assert ortho_im.crs == dem_crs
            assert ortho_im.res == (res, res)


def test_odm_dataset_dir_error(tmp_path: Path, runner: CliRunner):
    """Test ``oty odm`` raises an error with a non-existing --dataset-dir."""
    dataset_dir = 'unknown'
    cli_str = f'odm --dataset-dir {dataset_dir}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        '--dataset-dir' in result.stdout
        and 'directory' in result.stdout
        and dataset_dir in result.stdout
    )


def test_rpc_image(rpc_image_file: Path, ngi_dem_file: Path, tmp_path: Path, runner: CliRunner):
    """Test ``oty rpc`` with a source image that has RPC metadata."""
    cli_str = f'rpc --dem {ngi_dem_file} --out-dir {tmp_path} {rpc_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout

    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1
    with rio.open(ortho_files[0], 'r') as im:
        assert im.crs == rio.CRS.from_epsg(4979)


def test_rpc_image_no_rpc_error(
    ngi_image_file: Path, ngi_dem_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty rpc`` raises an error with a source image that has no RPC metadata."""
    cli_str = f'rpc --dem {ngi_dem_file} --out-dir {tmp_path} {ngi_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        'SOURCE' in result.stdout
        and ngi_image_file.name in result.stdout
        and 'No RPC parameters' in result.stdout
    )


def test_rpc_image_not_found_error(ngi_dem_file: Path, tmp_path: Path, runner: CliRunner):
    """Test ``oty rpc`` raises an error with a non-existent source image."""
    src_file = 'unknown.tif'
    cli_str = f'rpc --dem {ngi_dem_file} --out-dir {tmp_path} {src_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        'SOURCE' in result.stdout and 'No such file' in result.stdout and src_file in result.stdout
    )


def test_rpc_param_file(
    rpc_image_file: Path,
    ngi_dem_file: Path,
    rpc_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty rpc`` with ``--rpc-param`` uses the RPC parameter file."""
    # create source image with no RPCs (and same name as rpc_image_file)
    src_file = tmp_path.joinpath(rpc_image_file.name)
    with rio.open(rpc_image_file, 'r') as rpc_im:
        profile = rpc_im.profile
        profile.pop('rpcs', None)  # make sure no rpcs
        with rio.open(src_file, 'w', **profile) as no_rpc_im:
            no_rpc_im.write(rpc_im.read())

    # orthorectify
    cli_str = (
        f'rpc --dem {ngi_dem_file} --rpc-param {rpc_param_file} --out-dir {tmp_path} {src_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout

    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1
    with rio.open(ortho_files[0], 'r') as im:
        assert im.crs == rio.CRS.from_epsg(4979)


def test_rpc_param_file_invalid_error(
    rpc_image_file: Path,
    ngi_dem_file: Path,
    ngi_oty_int_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty rpc`` with ``--rpc-param`` raises an error with an invalid parameter file."""
    cli_str = (
        f'rpc --dem {ngi_dem_file} --rpc-param {ngi_oty_int_param_file} --out-dir {tmp_path}'
        f' {rpc_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        '--rpc-param' in result.stdout
        and 'valid' in result.stdout
        and ngi_oty_int_param_file.name in result.stdout
    )


def test_rpc_param_file_not_found_error(
    rpc_image_file: Path, ngi_dem_file: Path, tmp_path: Path, runner: CliRunner
):
    """Test ``oty rpc`` with ``--rpc-param`` raises an error with a non-existent parameter file."""
    rpc_param_file = 'unknown.yaml'
    cli_str = (
        f'rpc --dem {ngi_dem_file} --rpc-param {rpc_param_file} --out-dir {tmp_path}'
        f' {rpc_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code != 0, result.stdout
    assert (
        '--rpc-param' in result.stdout
        and 'No such file' in result.stdout
        and rpc_param_file in result.stdout
    )


def test_rpc_image_crs(rpc_image_file: Path, ngi_dem_file: Path, tmp_path: Path, runner: CliRunner):
    """Test ``oty rpc`` with ``--crs`` and a source image with RPC metadata, generates orthos with
    the correct CRS.
    """
    crs = 'EPSG:3857'
    cli_str = f'rpc --dem {ngi_dem_file} --crs {crs} --res 30 --out-dir {tmp_path} {rpc_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.crs == rio.CRS.from_string(crs)


def test_rpc_param_file_crs(
    rpc_image_file: Path,
    ngi_dem_file: Path,
    rpc_param_file: Path,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test ``oty rpc`` with ``--crs`` and a ``--rpc-param`` RPC parameter file, generates orthos
    with the correct CRS.
    """
    crs = 'EPSG:3857'
    cli_str = (
        f'rpc --dem {ngi_dem_file} --rpc-param {rpc_param_file} --crs {crs} --res 30 '
        f'--out-dir {tmp_path} {rpc_image_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.crs == rio.CRS.from_string(crs)


def test_rpc_option(rpc_image_file: Path, ngi_dem_file: Path, tmp_path: Path, runner: CliRunner):
    """Test ``oty rpc`` passes through a ``--res`` option."""
    res = 0.0005  # degrees
    cli_str = f'rpc --dem {ngi_dem_file} --res {res} --out-dir {tmp_path} {rpc_image_file}'
    result = runner.invoke(cli, cli_str.split())
    assert result.exit_code == 0, result.stdout
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1

    with rio.open(ortho_files[0], 'r') as im:
        assert im.res == (res, res)


def test_simple_ortho(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_legacy_config_file: Path,
    ngi_legacy_csv_file: Path,
    tmp_path: Path,
):
    """Test legacy ``simple-ortho`` CLI with NGI data."""
    cli_str = (
        f'-od {tmp_path} -rc {ngi_legacy_config_file} {ngi_image_file} {ngi_dem_file} '
        f'{ngi_legacy_csv_file}'
    )
    simple_ortho(cli_str.split())
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1


def test_simple_ortho_write_conf(
    ngi_image_file: Path,
    ngi_dem_file: Path,
    ngi_legacy_config_file: Path,
    ngi_legacy_csv_file: Path,
    tmp_path: Path,
):
    """Test legacy ``simple-ortho -wc`` writes a config file."""
    conf_file = tmp_path.joinpath('test_config.yaml')
    cli_str = (
        f'-od {tmp_path} -rc {ngi_legacy_config_file} -wc {conf_file} {ngi_image_file} '
        f'{ngi_dem_file} {ngi_legacy_csv_file}'
    )
    simple_ortho(cli_str.split())
    assert conf_file.exists()
    assert yaml.safe_load(conf_file.read_text()) == yaml.safe_load(
        ngi_legacy_config_file.read_text()
    )


def test__frame_mult_camera(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    mult_int_param_dict: dict,
    mult_ext_param_dict: dict,
    utm34n_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test the _ortho backend with multiple camera ID interior & exterior parameters."""
    # note that this test, and those below, basically duplicate similarly named tests in
    # test_factory
    src_files = []
    for src_file_key in mult_ext_param_dict.keys():
        src_file = tmp_path.joinpath(src_file_key).with_suffix(rgb_byte_src_file.suffix)
        shutil.copy(rgb_byte_src_file, src_file)
        src_files.append(str(src_file))

    cameras = FrameCameras(mult_int_param_dict, mult_ext_param_dict)

    _ortho(
        src_files=fsspec.open_files(src_files, 'rb'),
        dem_file=fsspec.open(str(float_utm34n_dem_file), 'rb'),
        cameras=cameras,
        crs=rio.CRS.from_string(utm34n_crs),
        dem_band=1,
        export_params=False,
        out_dir=fsspec.open(str(tmp_path), 'wb'),
        overwrite=False,
    )
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == len(src_files)


def test__frame_single_camera(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    pinhole_int_param_dict: dict,
    xyz: tuple,
    opk: tuple,
    utm34n_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test the _ortho backend with an exterior parameter camera ID of ``None`` and a single set of
    interior parameters.
    """
    ext_param_dict = {rgb_byte_src_file.name: dict(xyz=xyz, opk=opk, camera=None)}
    cameras = FrameCameras(pinhole_int_param_dict, ext_param_dict)

    _ortho(
        src_files=(fsspec.open(str(rgb_byte_src_file), 'rb'),),
        dem_file=fsspec.open(str(float_utm34n_dem_file), 'rb'),
        cameras=cameras,
        crs=rio.CRS.from_string(utm34n_crs),
        dem_band=1,
        export_params=False,
        out_dir=fsspec.open(str(tmp_path), 'wb'),
        overwrite=False,
    )
    ortho_files = [*tmp_path.glob('*_ORTHO.tif')]
    assert len(ortho_files) == 1


def test__frame_filename_not_found_error(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    pinhole_int_param_dict: dict,
    xyz: tuple,
    opk: tuple,
    utm34n_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test the _ortho backend raises an error when there are no exterior parameters for the
    source image.
    """
    ext_param_dict = {'unknown.tif': dict(xyz=xyz, opk=opk, camera='pinhole test camera')}
    cameras = FrameCameras(pinhole_int_param_dict, ext_param_dict)

    with pytest.raises(click.UsageError) as ex:
        _ortho(
            src_files=(fsspec.open(str(rgb_byte_src_file), 'rb'),),
            dem_file=fsspec.open(str(float_utm34n_dem_file), 'rb'),
            cameras=cameras,
            crs=rio.CRS.from_string(utm34n_crs),
            dem_band=1,
            export_params=False,
            out_dir=fsspec.open(str(tmp_path)),
            overwrite=False,
        )
    assert rgb_byte_src_file.name in str(ex.value) and 'exterior parameters' in str(ex.value)


def test__frame_camera_not_found_error(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    mult_int_param_dict: dict,
    xyz: tuple,
    opk: tuple,
    utm34n_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test the _ortho backend raises an error when there are no interior parameters for the
    exterior parameter camera ID.
    """
    camera = 'other'
    ext_param_dict = {rgb_byte_src_file.name: dict(xyz=xyz, opk=opk, camera=camera)}
    cameras = FrameCameras(mult_int_param_dict, ext_param_dict)

    with pytest.raises(click.UsageError) as ex:
        _ortho(
            src_files=(fsspec.open(str(rgb_byte_src_file), 'rb'),),
            dem_file=fsspec.open(str(float_utm34n_dem_file), 'rb'),
            cameras=cameras,
            crs=rio.CRS.from_string(utm34n_crs),
            dem_band=1,
            alpha=1.0,
            export_params=False,
            out_dir=fsspec.open(str(tmp_path)),
            overwrite=False,
            full_remap=True,
        )
    assert camera in str(ex.value) and 'interior parameters' in str(ex.value)


def test__frame_mult_camera_no_camera_error(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    mult_int_param_dict: dict,
    xyz: tuple,
    opk: tuple,
    utm34n_crs: str,
    tmp_path: Path,
    runner: CliRunner,
):
    """Test the _ortho backend raises an error when the exterior parameter camera ID is ``None`` and
    there are multiple sets of interior parameters.
    """
    ext_param_dict = {rgb_byte_src_file.name: dict(xyz=xyz, opk=opk, camera=None)}
    cameras = FrameCameras(mult_int_param_dict, ext_param_dict)

    with pytest.raises(click.UsageError) as ex:
        _ortho(
            src_files=(fsspec.open(str(rgb_byte_src_file), 'rb'),),
            dem_file=fsspec.open(str(float_utm34n_dem_file), 'rb'),
            cameras=cameras,
            crs=rio.CRS.from_string(utm34n_crs),
            dem_band=1,
            alpha=1.0,
            export_params=False,
            out_dir=fsspec.open(str(tmp_path)),
            overwrite=False,
            full_remap=True,
        )
    assert (
        rgb_byte_src_file.name in str(ex.value) and 'exterior parameters' in str(ex.value).lower()
    )



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

"""Utility functions for internal use."""
from __future__ import annotations

import cProfile
import logging
import os
import posixpath
import pstats
import tracemalloc
import warnings
from contextlib import contextmanager, ExitStack
from io import IOBase
from os import PathLike
from pathlib import Path
from typing import IO, Iterable

import cv2
import fsspec
import numpy as np
import rasterio as rio
from fsspec.core import OpenFile

try:
    from fsspec.implementations.http import HTTPFileSystem
except ImportError:
    HTTPFileSystem = type('unknown', (), {})

from fsspec.implementations.local import LocalFileSystem
from rasterio.crs import CRS
from rasterio.errors import NotGeoreferencedWarning, RasterioIOError
from rasterio.io import DatasetReaderBase, DatasetWriter
from rasterio.windows import Window

from orthority.enums import Interp

logger = logging.getLogger(__name__)


@contextmanager
def suppress_no_georef():
    """Context manager to suppress rasterio's NotGeoreferencedWarning."""
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=NotGeoreferencedWarning)
        yield


def expand_window_to_grid(win: Window, expand_pixels: tuple[int, int] = (0, 0)) -> Window:
    """Expand rasterio window extents to the nearest whole numbers."""
    col_off, col_frac = np.divmod(win.col_off - expand_pixels[1], 1)
    row_off, row_frac = np.divmod(win.row_off - expand_pixels[0], 1)
    width = np.ceil(win.width + 2 * expand_pixels[1] + col_frac)
    height = np.ceil(win.height + 2 * expand_pixels[0] + row_frac)
    exp_win = Window(
        col_off.astype('int'), row_off.astype('int'), width.astype('int'), height.astype('int')
    )
    return exp_win


def nan_equals(a: np.ndarray | float, b: np.ndarray | float) -> np.ndarray:
    """Compare two numpy objects a & b, returning true where elements of both a & b are nan."""
    return (a == b) | (np.isnan(a) & np.isnan(b))


def distort_image(camera, image: np.ndarray, nodata=0, interp=Interp.nearest) -> np.ndarray:
    """Return a distorted image given a frame camera model and source image."""

    if not np.all(np.array(image.shape[-2:][::-1]) == camera.im_size):
        raise ValueError("'image' shape should be the same as the 'camera' image size.")

    # create (j, i) pixel coords for distorted image
    j_range = np.arange(0, camera.im_size[0])
    i_range = np.arange(0, camera.im_size[1])
    j_grid, i_grid = np.meshgrid(j_range, i_range, indexing='xy')
    ji = np.array((j_grid.reshape(-1), i_grid.reshape(-1)))

    # find the corresponding undistorted/ source (j, i) pixel coords
    camera_xyz = camera._pixel_to_camera(ji)
    undist_ji = camera._K_undistort.dot(camera_xyz)[:2].astype('float32')

    def distort_band(src_array: np.ndarray, dst_array: np.ndarray):
        """Distort a 2D band array."""
        cv2.remap(
            src_array,
            undist_ji[0].reshape(image.shape[-2:]),
            undist_ji[1].reshape(image.shape[-2:]),
            Interp[interp].to_cv(),
            dst=dst_array,
            borderMode=cv2.BORDER_TRANSPARENT,
        )

    dist_image = np.full(image.shape, dtype=image.dtype, fill_value=nodata)
    for bi in range(image.shape[0]):
        distort_band(image[bi], dist_image[bi])

    return dist_image


@contextmanager
def profiler():
    """Context manager for profiling in DEBUG log level."""
    if logger.getEffectiveLevel() <= logging.DEBUG:
        proc_profile = cProfile.Profile()
        tracemalloc.start()
        proc_profile.enable()

        yield

        proc_profile.disable()
        # tottime is the total time spent in the function alone. cumtime is the total time spent
        # in the function plus all functions that this function called
        proc_stats = pstats.Stats(proc_profile).sort_stats('cumtime')
        logger.debug(f'Processing times:')
        proc_stats.print_stats(20)
        current, peak = tracemalloc.get_traced_memory()
        tracemalloc.stop()
        logger.debug(
            f"Memory usage: current: {current / 10 ** 6:.1f} MB, peak: {peak / 10 ** 6:.1f} MB"
        )
    else:
        yield


def utm_crs_from_latlon(lat: float, lon: float) -> CRS:
    """Return a 2D rasterio UTM CRS for the given (lat, lon) coordinates in degrees."""
    # adapted from https://gis.stackexchange.com/questions/269518/auto-select-suitable-utm-zone-based-on-grid-intersection
    zone = int(np.floor((lon + 180) / 6) % 60) + 1
    epsg = 32600 + zone if lat >= 0 else 32700 + zone
    return CRS.from_epsg(epsg)


def validate_collection(template: Iterable, coll: Iterable):
    """
    Validate a nested dict / list of values (``coll``) against a nested dict / list of types, tuples
    of types, and values (``template``).

    All items in a ``coll`` list are validated against the first item in the corresponding
    ``template`` list.
    """
    # adapted from https://stackoverflow.com/questions/45812387/how-to-validate-structure-or-schema-of-dictionary-in-python
    if isinstance(template, dict) and isinstance(coll, dict):
        # struct is a dict of types or other dicts
        for k in template:
            if k in coll:
                validate_collection(template[k], coll[k])
            else:
                raise KeyError(f"No key: '{k}'.")
    elif isinstance(template, list) and isinstance(coll, list) and len(template) and len(coll):
        # struct is list in the form [type or dict]
        for item in coll:
            validate_collection(template[0], item)
    elif isinstance(template, type):
        # struct is the type of conf
        if not isinstance(coll, template):
            raise TypeError(f"'{coll}' is not an instance of {template}.")
    elif isinstance(template, tuple) and all([isinstance(item, type) for item in template]):
        # struct is a tuple of types
        if not isinstance(coll, template):
            raise TypeError(f"'{coll}' is not an instance of any of {template}.")
    elif isinstance(template, object) and template is not None:
        # struct is the value of conf
        if not coll == template:
            raise ValueError(f"'{coll}' does not equal '{template}'.")


def get_filename(file: str | PathLike | OpenFile | DatasetReaderBase | IO) -> str:
    """Return a source filename for the given ``file`` object.  If ``file`` is an
    :class:`~fsspec.core.OpenFile` instance, a :class:`~rasterio.io.DatasetReaderBase` instance
    or file object, it should have a ``filename`` attribute i.e. have been created by either
    :class:`Open` or :class:`OpenRaster`.
    """
    if isinstance(file, DatasetReaderBase):
        filename = getattr(file, 'filename', Path(file.name).name)
    elif isinstance(file, OpenFile):
        filename = getattr(file, 'filename', Path(file.path).name)
    elif isinstance(file, IOBase):
        filename = getattr(file, 'filename', Path(getattr(file, 'name', '<file object>')).name)
    else:
        filename = Path(os.fspath(file)).name
    return filename


def join_ofile(base: str | PathLike | OpenFile, rel: str, mode: str = None, **kwargs) -> OpenFile:
    """Return an fsspec OpenFile whose path is a join of the ``base`` path with the ``rel`` path."""
    if not isinstance(base, OpenFile):
        base = fsspec.open(os.fspath(base), mode or 'rt')

    joined_path = posixpath.join(base.path, rel)
    return OpenFile(base.fs, joined_path, mode=mode or base.mode, **kwargs)


class OpenRaster:
    """
    Context manager for local or remote Rasterio datasets.

    :param file:
        A path, URI, :class:`~fsspec.core.OpenFile` instance, or open dataset.  If it is an open
        dataset, it is returned unaltered on entering the context, not closed on exiting the
        context, and ``mode`` and ``kwargs`` are ignored.  If is an OpenFile instance, it should
        be open in a binary mode matching ``mode``.
    :param mode:
        Mode in which the dataset is opened.  Either ``'r'`` or ``'w'``.
    :param overwrite:
        Whether to overwrite an existing file in ``'w'`` mode.  Ignored in ``'r'`` mode.
    :param kwargs:
        Keyword arguments to pass to :func:`rasterio.open`.
    """

    def __init__(
        self,
        file: str | PathLike | DatasetReaderBase | OpenFile,
        mode: str = 'r',
        overwrite: bool = False,
        **kwargs,
    ):
        if mode not in ['r', 'w']:
            raise ValueError(f"The 'mode' argument should be either 'r' or 'w', not '{mode}'.")

        self._exit_stack = ExitStack()

        if isinstance(file, DatasetReaderBase):
            if file.closed:
                raise IOError('Dataset is closed.')
            if mode not in file.mode:
                raise IOError(
                    f"Dataset mode: '{file.mode}' not compatible with the mode argument: '{mode}'."
                )
            self._dataset = file

        elif isinstance(file, (str, PathLike, OpenFile)):
            # TODO: use the opener arg to rio.open() and pin the rasterio dependency version when
            #  rasterio 1.4 is released, rather than passing an open file object.  this should
            #  allow sidecar files to read / written (test that it does).  also test that files
            #  are not buffered in memory with this option, and currently problematic fsspec
            #  protocols (e.g. github) no longer cause a seg fault.
            if isinstance(file, OpenFile):
                if mode + 'b' != file.mode:
                    raise IOError(
                        f"OpenFile object mode: '{file.mode}' should be a binary mode matching the "
                        f"mode argument: '{mode}'."
                    )
                ofile = file
            else:
                ofile = fsspec.open(os.fspath(file), mode + 'b')

            # TODO: delete sidecar files if overwriting
            if not overwrite and 'w' in mode and ofile.fs.exists(ofile.path):
                raise FileExistsError(f"File exists: '{ofile.path}'")

            if isinstance(ofile.fs, (LocalFileSystem, HTTPFileSystem)):
                # use GDAL internal file system
                try:
                    self._dataset = self._exit_stack.enter_context(
                        rio.open(ofile.path, mode, **kwargs)
                    )
                except RasterioIOError as ex:
                    raise FileNotFoundError(str(ex))
            else:
                # use fsspec file object
                file_obj = self._exit_stack.enter_context(ofile)
                self._dataset = self._exit_stack.enter_context(rio.open(file_obj, mode, **kwargs))

            # store the source filename as a dataset attribute
            self._dataset.filename = get_filename(file)

        else:
            raise TypeError(f"Unsupported 'file' type: {type(file)}")

    def __enter__(self) -> rio.DatasetReader | DatasetWriter:
        if self._dataset.closed:
            raise IOError('Dataset is closed.')
        return self._dataset

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._exit_stack.__exit__(exc_type, exc_val, exc_tb)

    def close(self):
        self._exit_stack.close()


class Open:
    """
    Context manager for local or remote file IO.

    :param file:
        A path, URI, :class:`~fsspec.core.OpenFile` instance, or file object.  If it is a file
        object, it is returned unaltered on entering the context, not closed on exiting the
        context, and ``mode`` and ``kwargs`` are ignored.  If is an OpenFile instance, it should
        be opened in ``mode`` (``kwargs`` are ignored).
    :param mode:
        Mode in which the file is opened.
    :param overwrite:
        Whether to overwrite an existing file in `'`w*'`` mode.  Ignored in ``'r*'`` mode.
    :param kwargs:
        Keyword arguments to pass to :func:`fsspec.open` or ``opener`` if it is specified.
    """

    def __init__(
        self,
        file: str | PathLike | IO | OpenFile,
        mode='rt',
        overwrite: bool = False,
        **kwargs,
    ):
        # TODO: text encoding defaults to utf-8, which can't be changed from the CLI
        self._exit_stack = ExitStack()
        if isinstance(file, IOBase):
            if file.closed:
                raise IOError('File object is closed.')
            if getattr(file, 'mode', mode) != mode:
                # note: fsspec text mode file objects do not have a mode property
                raise IOError(f"File object mode should match the mode argument: '{mode}'.")
            self._file_obj = file

        elif isinstance(file, (OpenFile, str, PathLike)):
            if isinstance(file, OpenFile):
                if mode != file.mode:
                    raise IOError(
                        f"OpenFile object mode: '{file.mode}', should match the mode argument:"
                        f" '{mode}'."
                    )
                ofile = file
            else:
                ofile = fsspec.open(os.fspath(file), mode, **kwargs)

            # overwrite could be prevented with 'x' modes, but is done this way for consistency
            # with OpenRaster & rasterio which doesn't support 'x'
            if not overwrite and 'w' in mode and ofile.fs.exists(ofile.path):
                raise FileExistsError(f"File exists: '{ofile.path}'")

            self._file_obj = self._exit_stack.enter_context(ofile)
            self._file_obj.filename = get_filename(file)

        else:
            raise TypeError(f"Unsupported 'file' type: {type(file)}")

    def __enter__(self) -> IO:
        return self._file_obj

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._exit_stack.__exit__(exc_type, exc_val, exc_tb)

    def close(self):
        self._exit_stack.close()



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

"""Orthorectification toolkit."""
import os

# enable on-demand download and caching of proj transformation grids (NB must be done before
# importing rasterio)
os.environ.update(PROJ_NETWORK='ON')

import logging
import pathlib

from orthority.enums import Compress, Interp
from orthority.factory import FrameCameras, RpcCameras
from orthority.ortho import Ortho


# Add a NullHandler to the package logger to hide logs by default.  Applications can then add
# their own handler(s).
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

# path to package root TODO: remove with deprecated simple-ortho CLI
if '__file__' in globals():
    root_path = pathlib.Path(__file__).absolute().parents[1]
else:
    root_path = pathlib.Path.cwd().absolute()
